#!/usr/bin/env python3
# -*- coding: utf-8, vim: expandtab:ts=4 -*-

from xtsv import build_pipeline, parser_skeleton, jnius_config, add_bool_arg


def main():

    argparser = parser_skeleton(description='EmPhon - a phonetic transcriber module for xtsv')

    add_bool_arg(argparser, 'ipaize',
                 ('Whether the output should be IPA or the emPhon inner representation, '
                  'which marks one phone with exactly one letter.'),
                 default=True, has_negative_variant=True)

    add_bool_arg(
        argparser, 'opt-palatal-assim',
        ('Whether optional palatal assimilation should happen with t/d+ny, e.g. lapátnyél -> lapátynyél'),
        default=False, has_negative_variant=True)

    add_bool_arg(
        argparser, 'include-sentence',
        'If on, there is a line of comment before the sentence that contains the entire surface form of the sentence.',
        default=True, has_negative_variant=True)

    opts = argparser.parse_args()

    jnius_config.classpath_show_warning = opts.verbose  # Suppress warning.

    # Set input and output iterators...
    if opts.input_text is not None:
        input_data = opts.input_text
    else:
        input_data = opts.input_stream
    output_iterator = opts.output_stream

    # Set the tagger name as in the tools dictionary
    used_tools = ['emphon']
    presets = []

    # Init and run the module as it were in xtsv

    emphon = ('emphon', 'EmPhon', 'EmPhon', (),
                        {'source_fields': {'form', 'anas'},
                         'target_fields': ['phon'],
                         'include_sentence': opts.include_sentence,
                         'transcriber_opts': {'ipaize': opts.ipaize,
                                              'optional_palatal_assimilation': opts.opt_palatal_assim},
                         },
              )

    emphon_ipa_comments = ('emphon', 'EmPhon', 'emPhon phonetic transcriber with IPAization and with comment lines', (),
                           {'source_fields': {'form', 'anas'},
                            'target_fields': ['phon'],
                            'include_sentence': True,
                            'transcriber_opts': {'ipaize': True,
                                                 'optional_palatal_assimilation': False},
                            },
                           )

    emphon_noipa_comments = ('emphon', 'EmPhon',
                             'emPhon phonetic transcriber without IPAization but with comment lines', (),
                             {'source_fields': {'form', 'anas'},
                              'target_fields': ['phon'],
                              'include_sentence': True,
                              'transcriber_opts': {'ipaize': False, 'optional_palatal_assimilation': False}, },)

    emphon_ipa_nocomments = ('emphon', 'EmPhon',
                             'emPhon phonetic transcriber with IPAization but without comment lines', (),
                             {'source_fields': {'form', 'anas'},
                              'target_fields': ['phon'],
                              'include_sentence': False,
                              'transcriber_opts': {'ipaize': True, 'optional_palatal_assimilation': False}, },)

    emphon_noipa_nocomments = (
        'emphon', 'EmPhon', 'emPhon phonetic transcriber without IPAization and comment lines', (),
        {'source_fields': {'form', 'anas'},
         'target_fields': ['phon'],
         'include_sentence': False, 'transcriber_opts': {'ipaize': False, 'optional_palatal_assimilation': False}, },)

    tools = [(emphon, ('emphon', 'emPhon phonetic transcriber ', 'emPhon'))]

    available_tools = [
        (emphon_ipa_comments, ('emphon-ipa-comments', 'emPhon-ipa-comments', 'emPhon-IPA-comments')),
        (emphon_ipa_nocomments, ('emphon-ipa-nocomments', 'emPhon-ipa-nocomments', 'emPhon-IPA-nocomments')),
        (emphon_noipa_comments, ('emphon-noipa-comments', 'emPhon-noipa-comments', 'emPhon-noIPA-comments')),
        (emphon_noipa_nocomments, ('emphon-noipa-nocomments', 'emPhon-noipa-nocomments', 'emPhon-noIPA-nocomments'))]

    # Run the pipeline on input and write result to the output...
    output_iterator.writelines(build_pipeline(input_data, used_tools, tools, presets, opts.conllu_comments))

    # TODO this method is recommended when debugging the tool
    # Alternative: Run specific tool for input (still in emtsv format):
    # from xtsv import process
    # from emphon import EmPhon
    # output_iterator.writelines(process(input_data, EmPhon(*emphon[3], **emphon[4])))

    # Alternative2: Run REST API debug server
    # from xtsv import pipeline_rest_api, singleton_store_factory
    # app = pipeline_rest_api('TEST', tools, {},  conll_comments=False, singleton_store=singleton_store_factory(),
    #                         form_title='TEST TITLE', doc_link='https://github.com/dlt-rilmta/emdummy')
    # app.run()


if __name__ == '__main__':
    main()



import re
import json
import os
from typing import List

MAIN_POS = {'/V', '/N', '/Adj', '/Adv'}


class Transcriber:
    """
    Phonetic transcriber class. Examples are commented after the rules to ease the reading of regexes.
    """
    def __init__(self, ipaize: bool = True, optional_palatal_assimilation: bool = False):
        """
        :param ipaize: whether the output uses the inner representation or the IPA form
        :param optional_palatal_assimilation: optional palatal assimilation in cases like `lapátnyél`
        """
        self.optional_palatal_assimilation = optional_palatal_assimilation

        self.ipaize = ipaize
        if self.ipaize:
            with open(os.path.join(os.path.dirname(__file__), 'ipa_key.json')) as infile:
                self.ipa_key = json.load(infile)

        self.double_letter_vocab = {
            'ccs': 'Č',
            'ddzs': 'Ĵ',
            'ddz': 'Ď',
            'ggy': 'Ǧ',
            'lly': 'J',
            'nny': 'Ɲ',
            'ssz': 'Ʃ',
            'tty': 'Ť',
            'zzs': 'Ž',
            'cs': 'č',
            'dzs': 'ĵ',
            'dz': 'ď',
            'gy': 'ǧ',
            'ly': 'j',
            'ny': 'ɲ',
            'sz': 'ʃ',
            'ty': 'ť',
            'zs': 'ž'}

    def __call__(self, sentence: str, passes=2) -> str:
        """
        Processes the incoming strings. Does `passes` passes so that a rule can feed into another rule.
        :param sentence: the sentence to process
        :param passes: the number of passes sentence goes under
        :return: the processed sentence
        """
        sentence = self.x_letter(sentence)
        sentence = self.double_letters(sentence)
        sentence = self.h_transformation(sentence)
        sentence = self.hiatus_filling(sentence)

        for _ in range(passes):

            sentence = self.n_nasalization(sentence)
            sentence = self.l_assimilation(sentence)
            sentence = self.degemination(sentence)
            sentence = self.n_assimilation(sentence)
            sentence = self.palatal_assimilation(sentence)
            sentence = self.m_nasalization(sentence)
            sentence = self.sibilant_assimilation(sentence)
            sentence = self.voice_assimilation(sentence)
            sentence = self.nasalisation(sentence)

        if self.ipaize:
            return self.ipaization(sentence)
        else:
            return sentence

    @staticmethod
    def x_letter(sentence: str) -> str:
        return re.sub(r'[Xx]', 'ksz', sentence)

    @staticmethod
    def _long_letters(sentence: str) -> str:
        sentence = re.sub(r'([bcdfghjklmnpqrstvxzčďǧɲʃťž])\1',
                          lambda m: m.group(1).upper(), sentence)
        return sentence

    @staticmethod
    def _stronger_long_letters(sentence: str) -> str:
        sentence = re.sub(r'([bcdfghjklmnpqrstvxzčďǧɲʃťž])[|~§#]?\1',
                          lambda m: m.group(1).upper(), sentence)
        return sentence

    def double_letters(self, sentence: str) -> str:

        sentence = re.sub(r'(ccs|ddzs|ddz|ggy|lly|nny|ssz|tty|zzs|cs|dzs|dz|gy|ly|ny|sz|ty|zs)',
                          lambda m: self.double_letter_vocab[m.group(1)], sentence)

        sentence = self._long_letters(sentence)

        return sentence

    def l_assimilation(self, sentence: str) -> str:
        sentence = re.sub(r'[lL][|§#~]?r', r'R', sentence)  # balra
        return self.double_letters(sentence)

    def h_transformation(self, sentence: str) -> str:
        sentence = re.sub(r'(ch$)', r'Ḧ', sentence)
        sentence = re.sub(
            r'([aáeéiíoóöőüűuú][|§#~ ]?)h([|§#~ ]?[aáeéiíoóöőüűuú])', r'\g<1>ɦ\g<2>', sentence)  # tehén
        sentence = re.sub(
            r'([aáeéiíoóöőüűuú][|§#~ ]?)H([|§#~ ]?[aáeéiíoóöőüűuú])', r'\g<1>Ḧ\g<2>', sentence)  # ahhoz
        sentence = re.sub(
            r'([aáeéiíoóöőüűuú][|§#~]?)[c]h([|§#~]?[bcdfgjklmnpqrstvxzčďǧɲʃťž ]?)', r'\g<1>ḧ\g<2>', sentence)  # pechből
        sentence = re.sub(r'([mnɲrlj][|§#~]?)h', r'\g<1>ɦ', sentence)

        return self._long_letters(sentence)

    def nasalisation(self, sentence: str) -> str:
        pairs = {'p': 'm', 'b': 'm', 'f': 'm', 'v': 'm', 'ǧ': 'ɲ', 'ť': 'ɲ'}
        sentence = re.sub(r'n([|~§#]?)([pbfvǧť])', lambda m: pairs[m.group(
            2)]+m.group(1)+m.group(2), sentence)  # tanpálya
        sentence = re.sub(r'(n)([|~§#]?ɲ)', r'Ɲ', sentence)  # lennyakkendő
        return self._long_letters(sentence)

    def n_assimilation(self, sentence: str) -> str:
        sentence = re.sub(r'n[|~§]?([lr])',
                          lambda m: m.group(1).upper(), sentence)  # hasonló
        return self._long_letters(sentence)

    def sibilant_assimilation(self, sentence: str) -> str:
        sentence = re.sub(r'(t)([|~§#]?ʃ)', r'C', sentence)  # hatszög
        sentence = re.sub(r'(t)([|~§#]?s)', r'Č', sentence)  # hátság
        sentence = re.sub(r'(t)([|~§# ]?c)', r'C', sentence)  # hét cica
        sentence = re.sub(r'(t)([|~§# ]?č)', r'Č', sentence)  # hat csap
        sentence = re.sub(r'(d)([|~§#]?ʃ)', r'C', sentence)  # rendszer
        sentence = re.sub(r'(d)([|~§#]?s)', r'Č', sentence)  # hadsereg
        return self._long_letters(sentence)

    def voice_assimilation(self, sentence: str) -> str:
        # voiced = 'bdǧgzžď'
        # voiceless = 'ptťkʃscf'
        pairs = {'p': 'b', 'b': 'p', 't': 'd', 'd': 't', 'ť': 'ǧ', 'ǧ': 'ť', 'k': 'g', 'g': 'k', 'f': 'v', 'v': 'f',
                 'ʃ': 'z', 'z': 'ʃ', 's': 'ž', 'ž': 's', 'c': 'ď', 'ď': 'c', 'h': 'ɦ', 'č': 'ĵ', 'ĵ': 'č'}

        sentence = re.sub(r'([bdǧgzžďvĵ])([|~§#]?[ptťkʃscfhč])',
                          lambda m: pairs[m.group(1)]+m.group(2), sentence)  # útpadka -> útpatka

        sentence = re.sub(r'([ptťkʃscfhč])([|~§#]?[bdǧgzžďĵ])',
                          lambda m: pairs[m.group(1)]+m.group(2), sentence)  # habfürdő -> hapfürdő

        return self._long_letters(sentence)

    def palatal_assimilation(self, sentence: str) -> str:
        # Rules in human-readable form
        # full = [
        #     ('ǧ', 'j', 'Ǧ'),
        #     ('d', 'j', 'Ǧ'),
        #     ('l', 'j', 'J'),
        #     ('n', 'j', 'Ɲ'),
        #     ('ɲ', 'j', 'Ɲ'),
        #     ('t', 'j', 'Ť'),
        #     ('ť', 'j', 'Ť')]
        # partial = [('d', 'ǧ', 'Ǧ'), ('t', 'ǧ', 'Ǧ'),
        #            ('d', 'ť', 'Ť'), ('t', 'ť', 'Ť')]
        # optional = [('d', 'ɲ', 'ǧɲ'), ('t', 'ɲ', 'ťɲ')]

        full_dict = {'ǧ': 'Ǧ', 'd': 'Ǧ', 'l': 'J',
                     'n': 'Ɲ', 'ɲ': 'Ɲ', 't': 'Ť', 'ť': 'Ť'}
        # partial is [dt][ǧť] and the second is upper
        optional_dict = {'d': 'ǧɲ', 't': 'ťɲ'}

        sentence = re.sub(r'([ǧdlnɲtť])[|~§#]?j',
                          lambda m: full_dict[m.group(1)], sentence)  # hagyjál

        sentence = re.sub(r'[dt][|~§# ]?([ǧť])',
                          lambda m: m.group(1).upper(), sentence)  # hét tyúk

        if self.optional_palatal_assimilation:
            sentence = re.sub(r'([dt])[|~§# ]?ɲ',
                              lambda m: optional_dict[m.group(1)], sentence)  # lapátnyél

        return self._long_letters(sentence)

    @staticmethod
    def hiatus_filling(sentence: str) -> str:
        sentence = re.sub(r'i[|~§]?([aáeéoóöőüűuú])', r'ij\g<1>', sentence)
        sentence = re.sub(r'([aáeéoóöőüűuú])[|~§]?i', r'\g<1>ji', sentence)
        return sentence

    def n_nasalization(self, sentence: str) -> str:
        sentence = re.sub(r'n[|~§#]?([gk])', r'ŋ\g<1>', sentence)
        return self.double_letters(sentence)

    def degemination(self, sentence: str) -> str:
        # These are left here intentionally, for reference.
        # consonants = 'bcdfghjklmnpqrstvwxzčďǧɲʃťž'
        # long_consonants = 'BCDFGHJKLMNPQRSTVWXZČĎǦƝƩŤŽ'
        # liquid = 'jlr'
        # nasal = 'mnɲ'
        # obstruents = 'bcdfghkpqstvwxzčďǧʃťž'

        # short consonant after long consonant
        sentence = re.sub(r'([BCDFGHJKLMNPQRSTVWXZČĎǦƝƩŤŽ])([|§~# ]?)([bcdfghjklmnpqrstvwxzčďǧɲʃťž])',
                          lambda m: m.group(1).lower()+m.group(2)+m.group(3), sentence)  # Ludd tábornok

        # short#1 short#2 short#2
        sentence = re.sub(r'([bcdfghjklmnpqrstvwxzčďǧɲʃťž]([bcdfghjklmnpqrstvwxzčďǧɲʃťž]))[|#§~ ]?\2',
                          r'\g<1>', sentence)  # Ford dísztárcsa

        # short long
        sentence = re.sub(r'([bcdfghjklmnpqrstvwxzčďǧɲʃťž])[|~#§]?([BCDFGHJKLMNPQRSTVWXZČĎǦƝƩŤŽ])',
                          lambda m: m.group(1)+m.group(2).lower(), sentence)  #

        # cons#1 cons#1 obstruent
        sentence = re.sub(
            r'([bcdfghjklmnpqrstvwxzčďǧɲʃťž])[|~#§ ]?(\1[bcdfghkpqstvwxzčďǧʃťž])', r'\g<2>', sentence)

        # long cons obstruent
        sentence = re.sub(r'([BCDFGHJKLMNPQRSTVWXZČĎǦƝƩŤŽ])([bcdfghkpqstvwxzčďǧʃťž])',
                          lambda m: m.group(1).lower()+m.group(2), sentence)

        # cons cons nasal
        sentence = re.sub(
            r'([bcdfghjklmnpqrstvwxzčďǧɲʃťž])[|~#§ ]?(\1[mnɲ])', r'\g<2>', sentence)  # optional

        # long cons nasal
        sentence = re.sub(r'([BCDFGHJKLMNPQRSTVWXZČĎǦƝƩŤŽ])([mnɲ])', lambda m: m.group(
            1).lower()+m.group(2), sentence)

        return self._long_letters(sentence)

    def m_nasalization(self, sentence: str) -> str:
        sentence = re.sub(r'[mn][|#§~ ]?([fv])', r'ɱ\g<1>', sentence)  # kámfor
        return self._long_letters(sentence)

    def ipaization(self, sentence: str) -> str:
        sentence = self._long_letters(re.sub(r'[|~§#]', '', sentence))
        ipa_sentence = ''.join(self.ipa_key.get(letter, letter) for letter in sentence)
        return ipa_sentence

    @staticmethod
    def segment(sentence: List[List[str]], field_names: dict) -> List[str]:
        """
        Segment words based on morphological analysis. It uses the field `anas`, which contains the output of emMorph
        with the possible segmentation. Selects the longest analysis (which is supposedly the finest), then adds
        delimiters based on the morph boundaries. In case of no analysis, it assumes that the token is a single morph.
        :param sentence: xtsv sentence with `form` and `morph` fields
        :param field_names: field names
        :return: list of segmented words
        """
        new_sentence = []

        for line in sentence:
            anas = json.loads(line[field_names['anas']])
            #  we suppose that the longest morphana is the finest
            anas_sorted = [ana['morphana']
                           for ana in sorted(anas, key=lambda s: s['morphana'].count('+'), reverse=True)]
            #  in case it is a symbol or some other weird thing the morphana is empty
            if anas_sorted == [] or anas_sorted == ['']:
                new_sentence.append(line[field_names['form']].lower() + ' ')
            else:
                #  the regex finds tag-surface realization pairs
                tags, morphs, _ = zip(*re.findall(r"\[(.*?)\]=(.*?)(\+|$)", anas_sorted[0]))
                morphs = [x.lower() for x in morphs]
                #  in some special cases like adverbs, punctiations there is only one morph and we have to consider it #
                #  a special case
                if len(morphs) == 1:
                    new_sentence.append(morphs[0] + ' ')
                else:
                    word = []
                    is_end_of_main = False
                    #  we iterate over the morphs and make 4 distinct categories:
                    #  before root, between roots, after root, between suffixes
                    #  we mark these by | # § ~
                    #  and for the empty morphs in the Hungarian at the end of words (Nom), we discard it
                    for i, morph in enumerate(morphs[:-1]):
                        word.append(morph)
                        if morphs[i+1] == '':
                            continue  # to avoid empty | in the end
                        curr_is_lemma, next_is_lemma = tags[i] in MAIN_POS, tags[i+1] in MAIN_POS
                        if is_end_of_main:
                            word.append('|')
                        elif curr_is_lemma and next_is_lemma:
                            word.append('#')
                        elif not curr_is_lemma and next_is_lemma:
                            word.append('§')
                        elif curr_is_lemma and not next_is_lemma:
                            word.append('~')
                            is_end_of_main = True
                        elif not curr_is_lemma and not next_is_lemma:
                            word.append('|')
                    else:
                        if morphs[-1] != '':  # empty morphemes at the end (nominative suffix) have to be discarded
                            word.append(morphs[-1])

                    new_sentence.append(''.join(word + [' ']))

        return new_sentence



from typing import List
from emphon.transcriber import Transcriber


class EmPhon:
    def __init__(
            self, source_fields=None, target_fields=None, transcriber_opts: dict = None, strict_xtsv_format=False,
            include_sentence=False):

        if source_fields is None:
            source_fields = set()

        if target_fields is None:
            target_fields = []

        if transcriber_opts is None:
            transcriber_opts = dict()

        self.source_fields = source_fields
        self.target_fields = target_fields

        self.transcriber = Transcriber(**transcriber_opts)

        self.include_sentence = include_sentence

    def comment_full_surface_form(self, segmented_sentence: List[str]) -> List[List[str]]:

        full_sentence = self.transcriber(''.join(segmented_sentence))

        return [['# phon = ' + full_sentence]]

    def process_sentence(self, sen, field_names=None):
        """
        Process one sentence per function call
        :param sen: The list of all tokens in the sentence, each token contain all fields
        :param field_names: The prepared field_names from prepare_fields() to select the appropriate input field
         to process
        :return: The sen object augmented with the output field values for each token
        """

        segmented_sentence = self.transcriber.segment(sentence=sen, field_names=field_names)

        per_word_sentence = [self.transcriber(word) for word in segmented_sentence]

        for line, transcription in zip(sen, per_word_sentence):
            line.append(transcription)

        if self.include_sentence:
            return self.comment_full_surface_form(segmented_sentence) + sen
        else:
            return sen

    @staticmethod
    def prepare_fields(field_names):
        """
        This function is called once before processing the input. It can be used to initialise field conversion classes
         to accomodate the current order of fields (eg. field to features)
        :param field_names: The dictionary of the names of the input fields mapped to their order in the input stream
        :return: The list of the initialised feature classes as required for process_sentence (in most cases the
         columnnumbers of the required field in the required order are sufficient
         e.g. return [field_names['form'], field_names['lemma'], field_names['xpostag'], ...] )
        """
        return field_names





import sys
import setuptools
import importlib.util


def import_python_file(module_name, file_path):
    # Import module from file: https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)
    return module


with open('README.md') as fh:
    long_description = fh.read()

setuptools.setup(
    name='emphon',
    # Get version without actually importing the module (else we need the dependencies installed)
    version=getattr(import_python_file('version', 'emphon/version.py'), '__version__'),
    author='levaid',  # Will warn about missing e-mail
    description='A phonetic transcriber module for emtsv',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/levaid/emPhon',
    # license='GNU Lesser General Public License v3 (LGPLv3)',  # Never really used in favour of classifiers
    # platforms='any',  # Never really used in favour of classifiers
    packages=setuptools.find_packages(include=['*', 'ipa_key.json'], exclude=['tests']),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Operating System :: POSIX :: Linux',
    ],
    python_requires='>=3.6',
    install_requires=['xtsv>=1.0.0,<2.0.0',
                      ],
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'emphon=emphon.__main__:main',
        ]
    },
)
